@@ -46,6 +46,7 @@
 #include <pwd.h>
 #include <errno.h>
 #include <getopt.h>
+#include <ctype.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -77,12 +78,128 @@
 #define GRST_PROXYCACHE		"/../proxycache/"
 #define GRST_BACKDATE_SECONDS	300
 
+#define END_ENTITY_ROBOT_OID "1.2.840.113612.5.2.3.3.1"
+
 static int GRSTx509CreateProxyRequest_int(char **reqtxt, char **keytxt, 
         char *ocspurl, int keysize);
 static int GRSTx509MakeProxyRequest_int(char **reqtxt, char *proxydir, 
         char *delegation_id, char *user_dn, int keysize);
 static int GRSTx509ProxyKeyMatch(char **pkfile, char *pkdir, STACK_OF(X509) *certstack); 
 
+static char *
+asn1_string2string(ASN1_STRING *str)
+{
+	BIO *bio;
+	int len, ret;
+	char *result = NULL;
+
+	bio = BIO_new(BIO_s_mem());
+	if (bio == NULL)
+		return NULL;
+
+	ASN1_STRING_print_ex(bio, str, ASN1_STRFLGS_RFC2253);
+
+	len = BIO_pending(bio);
+	if (len <= 0)
+		goto end;
+
+	result = calloc(1, len + 1);
+	if (result == NULL)
+		goto end;
+
+	ret = BIO_read(bio, result, len);
+	if (ret != len) {
+		free(result);
+		result = NULL;
+		goto end;
+	}
+
+end:
+	if (bio)
+		BIO_free(bio);
+
+	return result;
+}
+
+static int
+is_robot_subject(const char *p)
+{
+	if (strlen(p) <= 6)
+		return 0;
+
+	if (strncmp(p, "Robot", 5) != 0)
+		return 0;
+
+	if (('0' <= p[5] && p[5] <= '9') ||
+		('A' <= p[5] && p[5] <= 'Z') ||
+		('a' <= p[5] && p[5] <= 'z'))
+		return 0;
+
+	return 1;
+}
+
+static int
+is_robot_certificate(X509 *cert)
+{
+	int i, ret, found;
+	char *p;
+	char buf[64];
+	X509_NAME_ENTRY *ne;
+	X509_NAME *subject;
+	ASN1_STRING *value;
+	CERTIFICATEPOLICIES *policies = NULL;
+	POLICYINFO *policy;
+
+	subject = X509_get_subject_name(cert);
+	if (subject == NULL)
+		return 0;
+
+	found = 0;
+	i = -1;
+	while ((i = X509_NAME_get_index_by_NID(subject, NID_commonName, i)) >= 0) {
+		ne = X509_NAME_get_entry(subject, i);
+		value = X509_NAME_ENTRY_get_data(ne);
+		p = asn1_string2string(value);
+		if (p == NULL)
+			continue;
+		ret = is_robot_subject(p);
+		free(p);
+		if (ret == 1) {
+			found = 1;
+			break;
+		}
+	}
+	if (!found)
+		return 0;
+
+	policies = X509_get_ext_d2i(cert, NID_certificate_policies, &i, NULL);
+	if (policies == NULL)
+		return 0;
+	found = 0;
+	for (i = 0; i < sk_POLICYINFO_num(policies); i++) {
+		policy = sk_POLICYINFO_value(policies, i);
+		memset(buf, 0, sizeof(buf));
+		OBJ_obj2txt(buf, sizeof(buf), policy->policyid, 1);
+		if (strcmp(buf, END_ENTITY_ROBOT_OID) == 0) {
+			found = 1;
+			break;
+		}
+	}
+	if (!found)
+		return 0;
+
+	return 1;
+}
+
+static GRSTx509Cert *
+add_grst_cred(GRSTx509Cert *last_cred)
+{
+	GRSTx509Cert *cert;
+
+	cert = calloc(1, sizeof(*cert));
+	return cert;
+}
+
 int
 GRSTasn1FindField(const char *oid, char *coords,
         char *asn1string,
@@ -218,7 +335,7 @@ int GRSTx509ChainFree(GRSTx509Chain *chain)
 static int GRSTx509VerifySig(time_t *time1_time, time_t *time2_time,
                              unsigned char *txt, int txt_len,
                              unsigned char *sig, int sig_len, 
-                             X509 *cert, EVP_MD *md_type)
+                             X509 *cert, const EVP_MD *md_type)
 ///
 /// Returns GRST_RET_OK if signature is ok, other values if not.
 {   
@@ -278,16 +395,17 @@ static int GRSTx509VerifyVomsSig(time_t *time1_time, time_t *time2_time,
    int           ihash, isig, iinfo;
    char          *certpath, *certpath2, acvomsdn[200], dn_coords[200],
                   info_coords[200], sig_coords[200], hash_coords[200];
-   unsigned char *p;
+   const unsigned char *p;
    DIR           *vomsDIR, *vomsDIR2;
    struct dirent *vomsdirent, *vomsdirent2;
    X509          *cert;
    FILE          *fp;
-   EVP_MD	 *md_type = NULL;
+   const EVP_MD  *md_type = NULL;
    struct stat    statbuf;
    time_t         voms_service_time1 = GRST_MAX_TIME_T, voms_service_time2 = 0,
                   tmp_time1, tmp_time2;
    ASN1_OBJECT   *hash_obj = NULL;
+   int           ret;
 
    if ((vomsdir == NULL) || (vomsdir[0] == '\0')) return GRST_RET_FAILED;
 
@@ -332,7 +450,9 @@ static int GRSTx509VerifyVomsSig(time_t *time1_time, time_t *time2_time,
         {        
           if (vomsdirent->d_name[0] == '.') continue;
         
-          asprintf(&certpath, "%s/%s", vomsdir, vomsdirent->d_name);
+          ret = asprintf(&certpath, "%s/%s", vomsdir, vomsdirent->d_name);
+          if (ret == -1)
+            continue;
           stat(certpath, &statbuf);
           
           if (S_ISDIR(statbuf.st_mode))
@@ -348,8 +468,10 @@ static int GRSTx509VerifyVomsSig(time_t *time1_time, time_t *time2_time,
                 {
                   if (vomsdirent2->d_name[0] == '.') continue;
                   
-                  asprintf(&certpath2, "%s/%s/%s", 
-                           vomsdir, vomsdirent->d_name, vomsdirent2->d_name);
+                  ret = asprintf(&certpath2, "%s/%s/%s",
+                                 vomsdir, vomsdirent->d_name, vomsdirent2->d_name);
+                  if (ret == -1)
+                    continue;
                 
                   fp = fopen(certpath2, "r");
                   GRSTerrorLog(GRST_LOG_DEBUG, 
@@ -462,7 +584,7 @@ static int GRSTx509VerifyVomsSigCert(time_t *time1_time, time_t *time2_time,
    char          *lscpath, *p, *cacertpath,
                  *vodir, *vomscert_cadn, *vomscert_vomsdn, 
                  *lsc_cadn, *lsc_vomsdn;
-   unsigned char *q;
+   const unsigned char *q;
    unsigned long  issuerhash = 0;
    DIR           *voDIR;
    struct dirent *vodirent;
@@ -472,7 +594,7 @@ static int GRSTx509VerifyVomsSigCert(time_t *time1_time, time_t *time2_time,
    time_t         tmp_time;
    ASN1_OBJECT   *hash_obj = NULL;
    char		  coords[200];
-   EVP_MD        *md_type = NULL;
+   const EVP_MD  *md_type = NULL;
    time_t	  voms_service_time1 = 0, voms_service_time2 = GRST_MAX_TIME_T;
 
    if ((vomsdir == NULL) || (vomsdir[0] == '\0')) return GRST_RET_FAILED;
@@ -513,7 +635,9 @@ static int GRSTx509VerifyVomsSigCert(time_t *time1_time, time_t *time2_time,
    /* check issuer CA certificate */
 
    issuerhash = X509_NAME_hash(X509_get_issuer_name(vomscert));
-   asprintf(&cacertpath, "%s/%.8x.0", capath, issuerhash);
+   ret = asprintf(&cacertpath, "%s/%.8lx.0", capath, issuerhash);
+   if (ret == -1)
+      goto end;
 
    /* check voms cert DN matches DN from AC */
 
@@ -581,7 +705,9 @@ static int GRSTx509VerifyVomsSigCert(time_t *time1_time, time_t *time2_time,
      goto end;
    }
 
-   asprintf(&vodir, "%s/%s", vomsdir, voname);
+   ret = asprintf(&vodir, "%s/%s", vomsdir, voname);
+   if (ret == -1)
+     goto end;
    
    voDIR = opendir(vodir);
    if (voDIR == NULL) goto end;
@@ -600,7 +726,9 @@ static int GRSTx509VerifyVomsSigCert(time_t *time1_time, time_t *time2_time,
           
           if (strcmp(&(vodirent->d_name[strlen(vodirent->d_name)-4]), ".lsc") != 0) continue;
         
-          asprintf(&lscpath, "%s/%s", vodir, vodirent->d_name);
+          ret = asprintf(&lscpath, "%s/%s", vodir, vodirent->d_name);
+          if (ret == -1)
+            goto end;
           stat(lscpath, &statbuf);
 
           GRSTerrorLog(GRST_LOG_DEBUG, "Check LSC file %s for %s,%s",
@@ -798,8 +926,10 @@ static int GRSTx509ChainVomsAdd(GRSTx509Cert **grst_cert,
                  (j < taglist[itag].length);
                  ++j) ;
                                
-            asprintf(&voname, "%.*s", j-1, 
+            ret = asprintf(&voname, "%.*s", j-1,
                      &(asn1string[taglist[itag].start+taglist[itag].headerlength+1]));
+            if (ret == -1)
+                return GRST_RET_FAILED;
           }
 
         snprintf(vomscert_coords, sizeof(vomscert_coords), 
@@ -842,22 +972,29 @@ static int GRSTx509ChainVomsAdd(GRSTx509Cert **grst_cert,
 
              if (itag > -1)
                {
-                 (*grst_cert)->next = malloc(sizeof(GRSTx509Cert));
-                 *grst_cert = (*grst_cert)->next;
-                 bzero(*grst_cert, sizeof(GRSTx509Cert));
-               
-                 (*grst_cert)->notbefore = time1_time;
-                 (*grst_cert)->notafter  = time2_time;
-                 asprintf(&((*grst_cert)->value), "%.*s",
-                          taglist[itag].length,
-                          &asn1string[taglist[itag].start+
-                                      taglist[itag].headerlength]);
-                                      
-                 (*grst_cert)->errors = chain_errors; /* ie may be invalid */
-                 (*grst_cert)->type = GRST_CERT_TYPE_VOMS;
-                 (*grst_cert)->issuer = strdup(acvomsdn);
-                 (*grst_cert)->dn = strdup(user_cert->dn);
-		 (*grst_cert)->delegation = delegation;
+                 GRSTx509Cert *cred;
+
+                 cred = add_grst_cred(*grst_cert);
+                 if (cred == NULL)
+                     return GRST_RET_FAILED;
+
+                 cred->notbefore = time1_time;
+                 cred->notafter  = time2_time;
+                 ret = asprintf(&cred->value, "%.*s",
+                                taglist[itag].length,
+                                &asn1string[taglist[itag].start+
+                                taglist[itag].headerlength]);
+                 if (ret == -1) {
+                     free(cred);
+                     return GRST_RET_FAILED;
+                 }
+                 cred->errors = chain_errors; /* ie may be invalid */
+                 cred->type = GRST_CERT_TYPE_VOMS;
+                 cred->issuer = strdup(acvomsdn);
+                 cred->dn = strdup(user_cert->dn);
+                 cred->delegation = delegation;
+                 (*grst_cert)->next = cred;
+                 (*grst_cert) = cred;
                }
              else break;
            }
@@ -890,6 +1027,7 @@ int GRSTx509ChainLoad(GRSTx509Chain **chain,
    X509_EXTENSION *ex;
    time_t now;
    GRSTx509Cert *grst_cert, *new_grst_cert, *user_cert = NULL;
+   int is_robot = 0;
    
    GRSTerrorLog(GRST_LOG_DEBUG, "GRSTx509ChainLoadCheck() starts");
 
@@ -914,7 +1052,11 @@ int GRSTx509ChainLoad(GRSTx509Chain **chain,
    cert = sk_X509_value(certstack, depth - 1);
    subjecthash = X509_NAME_hash(X509_get_subject_name(cert));
    issuerhash = X509_NAME_hash(X509_get_issuer_name(cert));
-   asprintf(&cacertpath, "%s/%.8x.0", capath, issuerhash);
+   ret = asprintf(&cacertpath, "%s/%.8lx.0", capath, issuerhash);
+   if (ret == -1) {
+       *chain = NULL;
+       return GRST_RET_FAILED;
+   }
    
    GRSTerrorLog(GRST_LOG_DEBUG, "Look for CA root file %s", cacertpath);
 
@@ -1074,6 +1216,7 @@ int GRSTx509ChainLoad(GRSTx509Chain **chain,
                     user_cert = new_grst_cert;
                     new_grst_cert->delegation 
                        = (lastcert == NULL) ? i : i + 1;
+                    is_robot = is_robot_certificate(cert);
                   }
               } 
             else 
@@ -1150,6 +1293,18 @@ int GRSTx509ChainLoad(GRSTx509Chain **chain,
       } /* end of for loop */
 
    if (cacert != NULL) X509_free(cacert);
+
+   if (is_robot) {
+	   GRSTx509Cert *cred;
+
+	   cred = add_grst_cred(grst_cert);
+	   if (cred == NULL)
+		   return GRST_RET_FAILED;
+	   cred->type = GRST_CERT_TYPE_ROBOT;
+	   cred->dn = strdup(user_cert->dn);
+	   grst_cert->next = cred;
+	   grst_cert = cred;
+   }
  
    return GRST_RET_OK;
 }
@@ -1900,15 +2055,18 @@ char *GRSTx509CachedProxyFind(char *proxydir, char *delegation_id,
 {
   char *user_dn_enc, *proxyfile;
   struct stat statbuf;
+  int ret;
+
   if (!GRST_is_id_safe(delegation_id))    
     return NULL;
 
   user_dn_enc = GRSThttpUrlEncode(user_dn);
 
-  asprintf(&proxyfile, "%s/%s/%s/userproxy.pem",
-           proxydir, user_dn_enc, delegation_id);
-           
+  ret = asprintf(&proxyfile, "%s/%s/%s/userproxy.pem",
+                 proxydir, user_dn_enc, delegation_id);
   free(user_dn_enc);
+  if (ret == -1)
+    return NULL;
 
   if ((stat(proxyfile, &statbuf) != 0) || !S_ISREG(statbuf.st_mode))
     {
@@ -1939,8 +2097,12 @@ char *GRSTx509CachedProxyKeyFind(char *proxydir, char *delegation_id,
 
   user_dn_enc = GRSThttpUrlEncode(user_dn);
 
-  asprintf(&prvkeydir, "%s/cache/%s/%s/",
-          proxydir, user_dn_enc, delegation_id);
+  retval = asprintf(&prvkeydir, "%s/cache/%s/%s/",
+                    proxydir, user_dn_enc, delegation_id);
+  if (retval == -1) {
+    free(user_dn_enc);
+    return NULL;
+  }
   
   retval = GRSTx509ProxyKeyMatch(&prvkeyfilename, prvkeydir, certstack);
   free(prvkeydir);  
@@ -2000,7 +2162,7 @@ static int GRSTx509ProxyKeyMatch(char **pkfile, char *pkdir,
         ret = SSL_CTX_check_private_key(ssl_ctx);
         /* Success */
         if (ret == 1){
-            asprintf(pkfile, "%s", pk_file);
+            ret = asprintf(pkfile, "%s", pk_file);
             closedir(FD);
 	    free (pk_file);
 	    pk_file = NULL;
@@ -2031,7 +2193,7 @@ static void mkdir_printf(mode_t mode, char *fmt, ...)
   va_list ap;
   
   va_start(ap, fmt);
-  vasprintf(&path, fmt, ap);
+  ret = vasprintf(&path, fmt, ap);
   va_end(ap);
 
   ret = mkdir(path, mode);
@@ -2226,10 +2388,10 @@ int GRSTx509MakeProxyRequest_int(char **reqtxt, char *proxydir,
 
     /* make the new proxy private key */
 
-    asprintf(&prvkeyfile, "%s/cache/%s/%s/userkey_XXXXXX",
-            proxydir, user_dn_enc, delegation_id);
+    ret = asprintf(&prvkeyfile, "%s/cache/%s/%s/userkey_XXXXXX",
+                   proxydir, user_dn_enc, delegation_id);
 
-    if (prvkeyfile == NULL){
+    if (ret == -1 || prvkeyfile == NULL){
         free(user_dn_enc);
         return GRST_RET_FAILED;
     }
@@ -2334,7 +2496,7 @@ int GRSTx509ProxyDestroy(char *proxydir, char *delegation_id, char *user_dn)
 /// (Including GRST_RET_NO_SUCH_FILE if the private key or cert chain
 ///  were not found.)
 {
-  int              ret = GRST_RET_OK;
+  int              ret = GRST_RET_OK, res;
   char             *filename, *user_dn_enc;
 
   if (strcmp(user_dn, "cache") == 0) return GRST_RET_FAILED;
@@ -2346,10 +2508,10 @@ int GRSTx509ProxyDestroy(char *proxydir, char *delegation_id, char *user_dn)
 
   /* proxy file */
   
-  asprintf(&filename, "%s/%s/%s/userproxy.pem",
-           proxydir, user_dn_enc, delegation_id);
+  res = asprintf(&filename, "%s/%s/%s/userproxy.pem",
+                 proxydir, user_dn_enc, delegation_id);
 
-  if (filename == NULL)  
+  if (res == -1 || filename == NULL)
     {
       free(user_dn_enc);
       return GRST_RET_FAILED;
@@ -2360,10 +2522,10 @@ int GRSTx509ProxyDestroy(char *proxydir, char *delegation_id, char *user_dn)
 
   /* voms file */
   
-  asprintf(&filename, "%s/%s/%s/voms.attributes",
-           proxydir, user_dn_enc, delegation_id);
+  res = asprintf(&filename, "%s/%s/%s/voms.attributes",
+                 proxydir, user_dn_enc, delegation_id);
 
-  if (filename == NULL)  
+  if (res == -1 || filename == NULL)
     {
       free(user_dn_enc);
       return GRST_RET_FAILED;
@@ -2385,6 +2547,7 @@ int GRSTx509ProxyGetTimes(char *proxydir, char *delegation_id, char *user_dn,
   char  *docroot, *filename, *user_dn_enc;
   FILE  *fp;
   X509  *cert;
+  int ret;
 
   if (strcmp(user_dn, "cache") == 0) return GRST_RET_FAILED;
 
@@ -2393,12 +2556,11 @@ int GRSTx509ProxyGetTimes(char *proxydir, char *delegation_id, char *user_dn,
     
   user_dn_enc = GRSThttpUrlEncode(user_dn);
   
-  asprintf(&filename, "%s/%s/%s/userproxy.pem",
-           proxydir, user_dn_enc, delegation_id);
-           
+  ret = asprintf(&filename, "%s/%s/%s/userproxy.pem",
+                 proxydir, user_dn_enc, delegation_id);
   free(user_dn_enc);
-
-  if (filename == NULL) return GRST_RET_FAILED;
+  if (ret == -1 || filename == NULL)
+    return GRST_RET_FAILED;
 
   fp = fopen(filename, "r");
   free(filename);
@@ -2642,8 +2804,10 @@ int GRSTx509CacheProxy(char *proxydir, char *delegation_id,
     mkdir_printf(S_IRUSR | S_IWUSR | S_IXUSR, 
             "%s/%s/%s", proxydir, user_dn_enc, delegation_id);
 
-    asprintf(&proxyfile, "%s/%s/%s/userproxy.pem",
-            proxydir, user_dn_enc, delegation_id);
+    ret = asprintf(&proxyfile, "%s/%s/%s/userproxy.pem",
+                   proxydir, user_dn_enc, delegation_id);
+    if (ret == -1)
+        goto end;
     free(user_dn_enc);
     user_dn_enc = NULL;
 
